<?php

namespace Drupal\sigmaxim_data_entity\Plugin\Field\FieldWidget;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\monarch_data_entity\DataEntityInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\ConstraintViolationInterface;

/**
 * Defines the 'sigmaxim_simple_data_reference' field widget.
 *
 * @FieldWidget(
 *   id = "sigmaxim_simple_data_reference",
 *   label = @Translation("Data Reference"),
 *   field_types = {
 *     "entity_reference",
 *   },
 * )
 */
class SimpleDataReferenceWidget extends WidgetBase {

  /**
   * The entity_type.manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity_field.manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The token service.
   *
   * @var \Drupal\monarch_data_entity\TokenService
   */
  protected $tokenService;

  /**
   * The query cache.
   *
   * @var array
   */
  static protected $cache = [];

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $interface = parent::create($container, $configuration, $plugin_id, $plugin_definition);

    $interface->entityTypeManager = $container->get('entity_type.manager');
    $interface->entityFieldManager = $container->get('entity_field.manager');
    $interface->renderer = $container->get('renderer');
    $interface->tokenService = $container->get('monarch_data_entity.token_service');

    return $interface;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return parent::defaultSettings() + [
      'weights' => [
        'label' => [
          'weight' => -2,
        ],
        'id' => [
          'weight' => -1,
        ],
      ],
      'height' => 1,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element['height'] = [
      '#type' => 'number',
      '#title' => $this->t('Height'),
      '#description' => $this->t('Height of the select widgets. A height of 1 will use dropdown widgets instead.'),
      '#min' => 1,
      '#default_value' => max(1, $this->getSetting('height') ?? 0 ?: 0),
    ];

    $element['weights'] = [
      '#type' => 'table',
      '#header' => [
        $this->t('Field'),
        $this->t('Weight'),
        $this->t('Token'),
        $this->t('Enabled'),
      ],
      '#attributes' => [
        'id' => 'data-drill-down-field-order',
      ],
      '#tabledrag' => [
        [
          'action' => 'order',
          'relationship' => 'sibling',
          'group' => 'table-sort-weight',
        ],
      ],
    ];

    $weights = $this->getSetting('weights');
    $rows = [];

    $target_field_definitions = $this->entityFieldManager->getFieldDefinitions($this->fieldDefinition->getSetting('target_type'), current($this->fieldDefinition->getSetting('handler_settings')['target_bundles']));

    foreach ($target_field_definitions as $field_name => $target_field_definition) {
      $rows[$field_name] = [
        $target_field_definition->getLabel(),
        $weights[$field_name]['weight'] ?? 0,
        $weights[$field_name]['token'] ?? '',
        $weights[$field_name]['enabled'] ?? !($target_field_definition instanceof BaseFieldDefinition),
      ];
    }

    uasort($rows, function ($a, $b) {
      return intval($a[1]) - intval($b[1]);
    });

    foreach ($rows as $name => [$title, $weight, $token, $enabled]) {
      $element['weights'][$name] = [
        '#attributes' => [
          'class' => ['draggable'],
        ],
        '#weight' => $weight,
        'name' => [
          '#type' => 'inline_template',
          '#template' => $title . '<div style="color:#747474">ID: <em>' . $name . '</em></div>',
        ],
        'weight' => [
          '#type' => 'weight',
          '#title' => $this->t('Weight for @title', [
            '@title' => $title,
          ]),
          '#title_display' => 'invisible',
          '#default_value' => $weight,
          '#attributes' => [
            'class' => [
              'table-sort-weight',
            ],
          ],
        ],
        'token' => [
          '#type' => 'textfield',
          '#default_value' => $token,
        ],
        'enabled' => [
          '#type' => 'checkbox',
          '#default_value' => $enabled,
        ],
      ];
    }

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    return [];
  }

  /**
   * Get field list.
   */
  public function getFieldList() {
    $field_weights = $this->getSetting('weights') ?? [];

    foreach ($field_weights as $field_name => $field_config) {
      if (!(($field_config['enabled'] ?? NULL) || ($field_config['token'] ?? NULL))) {
        unset($field_weights[$field_name]);
      }
    }

    uasort($field_weights, function ($a, $b) {
      return intval($a['weight'] ?? 0) - intval($b['weight'] ?? 0);
    });

    $target_field_definitions = $this->entityFieldManager->getFieldDefinitions($this->fieldDefinition->getSetting('target_type'), current($this->fieldDefinition->getSetting('handler_settings')['target_bundles']));

    return empty($field_weights) ? array_keys($target_field_definitions) : array_keys($field_weights);
  }

  /**
   * Flatten with key path.
   */
  protected function flatten(array $value, string $path = NULL, array &$ret = []) {
    foreach ($value as $key => $subvalue) {
      $subpath = empty($path) ? $key : $path . '.' . $key;

      if (is_array($subvalue)) {
        $this->flatten($subvalue, $subpath, $ret);
      }
      else {
        $ret[$subpath] = $subvalue;
      }
    }

    return $ret;
  }

  /**
   * Get entity based on the input values.
   */
  protected function getEntityIdsFromInputs(array $input = [], int &$count = NULL, bool $staged = FALSE) : array {
    $target_fields = $this->getFieldList();
    $target_type = $this->fieldDefinition->getSetting('target_type');
    $es = $this->entityTypeManager->getStorage($target_type);

    $query_values = [];
    $query_values['bundle'] = current($this->fieldDefinition->getSetting('handler_settings')['target_bundles']);

    $all_query_values = $query_values;
    $last_value_field = NULL;

    foreach ($target_fields as $target_field_name) {
      if ($target_field_name === 'id') {
        $input_field_value = $input[$target_field_name][$target_field_name] ?? FALSE ?: FALSE;
      }
      else {
        $input_field_value = json_decode($input[$target_field_name][$target_field_name] ?? '' ?: '', TRUE);
      }

      $all_query_values[$target_field_name] = $input_field_value;

      if (!empty($input_field_value)) {
        $last_value_field = $target_field_name;
      }
    }

    $cache_key = 'getEntityIdFromInputs|' . $target_type . '|' . $staged . '|' . json_encode($all_query_values);

    if (isset(static::$cache[$cache_key])) {
      $res = static::$cache[$cache_key] ?? [];

      if ($staged) {
        $count = count(end($res) ?: []);
      }
      else {
        $count = count($res);
      }
    }
    else {
      $all_query = $es->getQuery()->accessCheck(FALSE);

      $res = [];

      if ($staged) {
        $found_last_value_field = is_null($last_value_field) ? 1 : 0;

        foreach ($this->flatten($all_query_values) as $key => $value) {
          $target_field_name = current(explode('.', $key));

          if ($target_field_name === $last_value_field) {
            $found_last_value_field = 1;
          }

          if ($found_last_value_field) {
            $found_last_value_field++;
          }

          if ($found_last_value_field <= 2) {
            if (empty($value)) {
              $res[$target_field_name] = end($res);
            }
            else {
              $all_query->condition($key, $value);
              $res[$target_field_name] = $all_query->execute() ?: [];

              if (count($res[$target_field_name]) < 2) {
                break;
              }
            }
          }
        }

        $count = count(end($res) ?: []);
        reset($res);
      }
      else {
        foreach (array_filter($this->flatten($all_query_values)) as $key => $value) {
          $all_query->condition($key, $value);
        }

        $res = $all_query->execute() ?: [];
        $count = count($res);
      }

      static::$cache[$cache_key] = $res;
    }

    return $res ?? [];
  }

  /**
   * Handles populating and updating the form, so Ajax can work properly.
   *
   * @param bool $isAjax
   *   Is this part of an ajax request?
   * @param \Drupal\Core\Field\FieldItemInterface $item
   *   The item that this element represents.
   * @param mixed $delta
   *   The field delta.
   * @param array $element
   *   The element render array.
   * @param array $form
   *   The entire form render array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param array $input
   *   The input to be used to populate the form. Leave blank to reset.
   *
   * @return void
   *   No return value.
   */
  public function updateOptions(bool $isAjax, FieldItemInterface $item = NULL, $delta, array &$element, array &$form, FormStateInterface $form_state, array $input = []) {
    $form_state->setProcessInput(TRUE);
    $field_name = $this->fieldDefinition->getName();

    $target_fields = $this->getFieldList();
    $target_field_definitions = $this->entityFieldManager->getFieldDefinitions($this->fieldDefinition->getSetting('target_type'), current($this->fieldDefinition->getSetting('handler_settings')['target_bundles']));
    $es = $this->entityTypeManager->getStorage($this->fieldDefinition->getSetting('target_type'));

    $parents = $form['#parents'];

    $id_prefix = implode('--', array_merge($parents, [
      $field_name,
      $delta,
      'wrapper',
    ]));
    $id_prefix = str_replace('_', '-', $id_prefix);
    $id_prefix = preg_replace("/[^0-9A-Za-z_\\-]+/", '-', $id_prefix);

    $element['#attached']['library'][] = 'monarch_data_entity/view_widget_notifier';
    $element['#prefix'] = "<div id=\"$id_prefix\">";
    $element['#suffix'] = '</div>';

    $ajax = [
      'callback' => [$this, 'ajax'],
      'disable-refocus' => FALSE,
      'event' => 'change',
      'wrapper' => $id_prefix,
      'progress' => [
        'type' => 'throbber',
      ],
    ];

    $orig_entity = $item->entity ?? NULL;

    foreach ($target_fields as $target_field_name) {
      $token = $this->settings['weights'][$target_field_name]['token'] ?? NULL ?: NULL;
      $enabled = $this->settings['weights'][$target_field_name]['enabled'] ?? FALSE ?: FALSE;
      $element[$target_field_name]['#type'] = 'html_tag';
      $element[$target_field_name]['#tag'] = 'div';

      $element[$target_field_name][$target_field_name]['#type'] = 'select';

      if (is_numeric($this->getSetting('height')) && $this->getSetting('height') > 0) {
        $element[$target_field_name][$target_field_name]['#size'] = $this->getSetting('height');
      }
      $element[$target_field_name][$target_field_name]['#options'] = [];

      if (!empty($orig_entity) && !empty($orig_entity->{$target_field_name}[$delta] ?? NULL) && !$orig_entity->{$target_field_name}[$delta]->isEmpty()) {
        $element[$target_field_name][$target_field_name]['#default_value'] = json_encode($orig_entity->{$target_field_name}[$delta]->getValue() ?? NULL);

        if (!$isAjax) {
          $input[$target_field_name][$target_field_name] = $element[$target_field_name][$target_field_name]['#default_value'];
        }
      }

      if (!$enabled) {
        $element[$target_field_name]['#attributes']['style'] = 'display:none;';
      }

      if ($token) {
        $element[$target_field_name][$target_field_name]['#after_build'][] = [
          $this,
          'afterBuildElement',
        ];
      }

      $element[$target_field_name][$target_field_name]['#ajax'] = $ajax;
    }

    if ($isAjax) {
      foreach ($target_fields as $target_field_name) {
        $form_state->setValue($element[$target_field_name][$target_field_name]['#parents'], $element[$target_field_name][$target_field_name]['#value']);
      }
    }

    $query_values = [];
    $query_values['bundle'] = current($this->fieldDefinition->getSetting('handler_settings')['target_bundles']);

    $entities = $es->loadByProperties([
      'bundle' => current($this->fieldDefinition->getSetting('handler_settings')['target_bundles']),
    ]);

    foreach ($target_fields as $target_field_name) {
      $element[$target_field_name]['#attributes']['style'] = 'display:inline-block';

      if (is_null($target_field_definitions[$target_field_name] ?? NULL)) {
        $orig_name = $this->fieldDefinition->getOriginalId();
        throw new \Exception("Field widget settings are out of date for \"$orig_name\". Please update the widget settings to make this widget work.");
      }

      $element[$target_field_name][$target_field_name]['#options'] = [
        '' => $this->t('-- Choose Value for @label --', [
          '@label' => $target_field_definitions[$target_field_name]->getLabel(),
        ]),
      ];

      foreach ($entities as $entity) {
        /** @var \Drupal\monarch_data_entity\DataEntityInterface $entity */
        if ($entity->hasField($target_field_name)) {
          $items = $entity->get($target_field_name);
          $element[$target_field_name][$target_field_name]['#dummy_items'] = $items::createInstance($items->getDataDefinition(), $items->getName());
          foreach ($items as $field_item) {
            if ($field_item->isEmpty()) {
              continue;
            }

            $evb = $this->entityTypeManager->getViewBuilder($this->fieldDefinition->getSetting('target_type'));
            $r = $evb->viewFieldItem($field_item, [
              'label_display' => 'none',
            ]);

            if ($target_field_name === 'id') {
              $element[$target_field_name][$target_field_name]['#options'][$field_item->value] = strip_tags($this->renderer->renderPlain($r));
            }
            else {
              $element[$target_field_name][$target_field_name]['#options'][json_encode($field_item->getValue())] = strip_tags($this->renderer->renderPlain($r));
            }
          }
        }
      }

      if (count($element[$target_field_name][$target_field_name]['#options'] ?? []) < 2) {
        unset($element[$target_field_name]);
      }

      if ($input[$target_field_name][$target_field_name] ?? NULL) {
        $input_field_value = $target_field_name !== 'id' ? json_decode($input[$target_field_name][$target_field_name], TRUE) : $input[$target_field_name][$target_field_name];
        $query_values[$target_field_name] = $input_field_value;
      }
    }

    $element['target-id-display'] = [
      '#type' => 'html_tag',
      '#tag' => 'div',
    ];

    $element['#after_build'][] = [
      $this,
      'renderResultingEntity',
    ];
  }

  /**
   * The '#after_build' callback.
   */
  public function afterBuildElement(array $element, FormStateInterface $form_state) {
    /** @var \Drupal\Core\Field\FieldItemListInterface $items */
    $items = $element['#dummy_items'];
    $target_field_name = $items->getName();
    $token = $this->settings['weights'][$target_field_name]['token'] ?? NULL ?: NULL;

    if ($token) {
      $form = $form_state->getCompleteForm();
      $value = $this->tokenService->transformArgumentToken($token, $element, NULL, $form, $form_state);
      $items->setValue($value);

      if ($items->isEmpty()) {
        $value = '';
      }
      elseif ($target_field_name === 'id') {
        $value = $items->value;
      }
      else {
        $value = json_encode(current($items->getValue()));
      }

      if ($form_state->isRebuilding()) {
        $form_state->setValue($element['#parents'], $value);
        $element['#value'] = $value;
      }
    }
    else {
      unset($element['#value']);
    }

    return $element;
  }

  /**
   * The '#after_build' callback.
   */
  public function renderResultingEntity(array $element, FormStateInterface $form_state) {
    $form = $form_state->getCompleteForm();
    $parents = [...$form['#parents'], $this->fieldDefinition->getName()];
    $delta = $element["#parents"][count($element['#parents']) - 2];
    $values = $form_state->getValue([...$parents, $delta, 'wrapper']) ?? NULL ?: [];
    $stages = $this->getEntityIdsFromInputs($values, $count, TRUE);
    $id = $count === 1 ? current(end($stages)) : NULL;
    $es = $this->entityTypeManager->getStorage($this->fieldDefinition->getSetting('target_type'));
    $display_entity = !empty($id) ? $es->load($id) : $id;

    $stage_keys = ['bundle', ...$this->getFieldList()];
    $prev_stage = reset($stage_keys);
    $current_stage = next($stage_keys);

    while ($current_stage) {
      $target_field_name = current(explode('.', $current_stage));
      $stage_entities = $es->loadMultiple($stages[$prev_stage] ?? []);
      $keep_values = [];

      if (count($stage_entities) > 1) {
        foreach ($stage_entities as $stage_entity) {
          if ($target_field_name === 'id') {
            $keep_values[$target_field_name] = $stage_entity->id();
          }
          elseif ($stage_entity instanceof FieldableEntityInterface && $stage_entity->hasField($target_field_name) && !$stage_entity->get($target_field_name)->isEmpty()) {
            foreach ($stage_entity->get($target_field_name) as $target_field_value) {
              /** @var \Drupal\Core\Field\FieldItemInterface $target_field_value */
              $keep_value = json_encode($target_field_value->getValue());
              $keep_values[$keep_value] = $keep_value;
            }
          }
        }
      }

      if (isset($element[$target_field_name][$target_field_name]['#options'])) {
        foreach ($element[$target_field_name][$target_field_name]['#options'] as $key => $value) {
          if (!empty($key) && is_null($keep_values[$key] ?? NULL)) {
            unset($element[$target_field_name][$target_field_name]['#options'][$key]);
          }
        }
      }

      if (count($keep_values) > 1) {
        $element[$target_field_name]['#attributes']['style'] = 'display:inline-block';
      }
      else {
        $element[$target_field_name]['#attributes']['style'] = 'display:none';
      }

      [$prev_stage, $current_stage] = [$current_stage, next($stage_keys)];
    }

    $element['target-id-display'] = [
      '#type' => 'html_tag',
      '#tag' => 'div',
    ];

    if ($count === 1 && ($display_entity = $es->load($id))) {
      $evb = $this->entityTypeManager->getViewBuilder($this->fieldDefinition->getSetting('target_type'));
      $element['target-id-display']['entity'] = $evb->view($display_entity);
    }

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    $element['#field_item'] = $items[$delta] ?? NULL;
    $element['wrapper'] = [];
    $element['wrapper']['#field_item'] = $items[$delta] ?? NULL;
    $parents = [...$form['#parents'], $this->fieldDefinition->getName()];
    $values = $form_state->getValue($parents)[$delta]['wrapper'] ?? [];

    $this->updateOptions(FALSE, $items[$delta], $delta, $element['wrapper'], $form, $form_state, $values);

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function form(FieldItemListInterface $items, array &$form, FormStateInterface $form_state, $get_delta = NULL) {
    $ret = parent::form($items, $form, $form_state, $get_delta);

    return $ret;
  }

  /**
   * Ajax callback.
   */
  public function ajax(array &$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    $elem = &$form;
    $elem_path = $triggering_element['#array_parents'];

    while (end($elem_path) !== 'wrapper') {
      array_pop($elem_path);
    }

    for ($key = reset($elem_path); $key !== FALSE; $key = next($elem_path)) {
      $tmp = &$elem[$key];
      $elem = &$tmp;
    }

    return $elem;
  }

  /**
   * {@inheritdoc}
   */
  public function errorElement(array $element, ConstraintViolationInterface $error, array $form, FormStateInterface $form_state) {
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {
    $parents = [...$form['#parents'], $this->fieldDefinition->getName()];
    $user_input = $form_state->getUserInput();
    $user_input = NestedArray::getValue($user_input, $parents);

    $fields = $this->getFieldList();
    unset($fields['id']);
    unset($fields['label']);
    $fields = array_fill_keys($fields, TRUE);

    foreach ($values as $delta => &$value) {
      $input = $value['wrapper'];

      foreach (array_keys($input) as $key) {
        if ($fields[$key] ?? NULL) {
          continue;
        }

        unset($input[$key]);
      }

      $ids = $this->getEntityIdsFromInputs($input, $count);
      $values[$delta]['target_id'] = $count === 1 ? current($ids) : NULL;
      unset($values[$delta]['wrapper']);
    }

    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    $target_type = $field_definition->getSetting('target_type');
    $entity_type = \Drupal::entityTypeManager()->getDefinition($target_type);

    if ($entity_type) {
      $target_bundles = $field_definition->getSetting('handler_settings')['target_bundles'] ?? NULL;

      if (!empty($target_bundles) && is_array($target_bundles) && count($target_bundles) == 1) {
        return $entity_type->entityClassImplements(DataEntityInterface::class);
      }
    }

    return FALSE;
  }

}
